/*! AutocompleteTextFieldExtension - v1.0.0 - 2016-02-02
* Copyright (c) 2016 Max Schuster; Licensed Apache-2.0 */
/* global define */

(function (root, factory) {
    var name = 'autoComplete';
    if (typeof define === 'function' && define.amd) {
        define([], function () {
            return (root[name] = factory());
        });
    } else if (typeof exports === 'object') {
        module.exports = factory();
    } else {
        root[name] = factory();
    }
}(this, function () {

    /* jshint validthis: true */

    "use strict";

    /**
     * @constructor
     * @param {Object} options
     * @returns {autoComplete}
     */
    function autoComplete(options) {
        // Check if requirements are met
        if (!document.querySelector) {
            return;
        }

        this.settings = this.buildSettings(options);
        this.eventHandlers = [];
        this.instances = this.createInstances(this.settings.selector);

        this.addEvent(window, 'resize', this.onResize, null);
    }

    autoComplete.prototype = {
        defaults: {
            selector: 0,
            source: 0,
            minChars: 3,
            delay: 150,
            cache: 1,
            menuClass: '',
            renderItem: function (item, search) {
                // escape special characters
                search = search.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&');
                var re = new RegExp("(" + search.split(' ').join('|') + ")", "gi");
                return '<div class="autocomplete-suggestion" data-val="' + item + '">' +
                        item.replace(re, "<b>$1</b>") +
                        '</div>';
            },
            onSelect: function (e, term, item) {},
            popupContainer: document.body
        },
        buildSettings: function (options) {
            var settings = {},
                    defaults = this.defaults;
            for (var property in defaults) {
                if (options.hasOwnProperty(property)) {
                    settings[property] = options[property];
                } else {
                    settings[property] = defaults[property];
                }
            }
            return settings;
        },
        createInstances: function (selector) {
            var textFields = typeof selector === 'object' ?
                    [selector] : document.querySelectorAll(selector),
                    instances = [],
                    settings = this.settings;

            for (var i = 0; i < textFields.length; i++) {
                var textField = textFields[i],
                        sc = this.createSuggestionsContainer(),
                        instance = {
                            textField: textField,
                            suggestionsContainer: sc,
                            autocompleteAttr: textField.getAttribute('autocomplete'),
                            cache: {},
                            lastVal: ''
                        },
                eventData = {
                    instance: instance
                };

                this.live('autocomplete-suggestion', 'mouseleave', this.onMouseLeave, sc, eventData);
                this.live('autocomplete-suggestion', 'mouseover', this.onMouseOver, sc, eventData);
                this.live('autocomplete-suggestion', 'mousedown', this.onMouseDown, sc, eventData);
                
                this.addEvent(textField, 'blur', this.onBlur, eventData);
                this.addEvent(textField, 'keydown', this.onKeyDown, eventData);
                this.addEvent(textField, 'keyup', this.onKeyUp, eventData);
                if (!settings.minChars) { 
                    this.addEvent(textField, 'focus', this.onFocus, eventData);
                }

                textField.autoCompleteInstance = instance;
                textField.setAttribute('autocomplete', 'off');

                settings.popupContainer.appendChild(sc);
                instances.push(instance);
            }

            return instances;
        },
        createSuggestionsContainer: function () {
            var container = document.createElement('div');
            container.className = 'autocomplete-suggestions ' + this.settings.menuClass;
            return container;
        },
        getStyle: function (element) {
            return window.getComputedStyle ?
                    getComputedStyle(element, null) : element.currentStyle;
        },
        isVisible: function (instance) {
            var sc = instance.suggestionsContainer;
            return this.getStyle(sc).display === "block";
        },
        hide: function (instance) {
            var sc = instance.suggestionsContainer;
            sc.style.display = 'none';
        },
        updateSuggestionsContainer: function (instance, resize, next) {
            var textField = instance.textField,
                    sc = instance.suggestionsContainer,
                    rect = textField.getBoundingClientRect(),
                    style = sc.style;
            style.left = rect.left + (window.pageXOffset || document.documentElement.scrollLeft) + 'px';
            style.top = rect.bottom + (window.pageYOffset || document.documentElement.scrollTop) + 1 + 'px';
            style.width = rect.right - rect.left + 'px'; // outerWidth
            if (!resize) {
                style.display = 'block';
                if (!sc.maxHeight) {
                    sc.maxHeight = parseInt(this.getStyle(sc).maxHeight);
                }
                if (!sc.suggestionHeight) {
                    sc.suggestionHeight = sc.querySelector('.autocomplete-suggestion').offsetHeight;
                }
                if (sc.suggestionHeight) {
                    if (!next) {
                        sc.scrollTop = 0;
                    } else {
                        var scrTop = sc.scrollTop,
                                selTop = next.getBoundingClientRect().top - sc.getBoundingClientRect().top;
                        if (selTop + sc.suggestionHeight - sc.maxHeight > 0) {
                            sc.scrollTop = selTop + sc.suggestionHeight + scrTop - sc.maxHeight;
                        } else if (selTop < 0) {
                            sc.scrollTop = selTop + scrTop;
                        }
                    }
                }
            }
        },
        suggest: function (instance, term, suggestions) {
            var sc = instance.suggestionsContainer,
                    settings = this.settings;
            if (settings.cache) {
                instance.cache[term] = suggestions;
            }
            if (suggestions.length && term.length >= settings.minChars) {
                var s = '';
                for (var i = 0; i < suggestions.length; i++) {
                    s += settings.renderItem(suggestions[i], term);
                }
                sc.innerHTML = s;
                this.updateSuggestionsContainer(instance, false);
            } else {
                this.hide(instance);
            }
        },
        hasClass: function (el, className) {
            if (el.classList) {
                return el.classList.contains(className);
            } else {
                return new RegExp('\\b' + className + '\\b').test(el.className);
            }
        },
        /**
         * Adds an event listener to the given element that calls the given hander.
         * 
         * @param {Element} el The element.
         * @param {String} type Event type. Can be a space separated list of types.
         * @param {Function} handler The handler function.
         * @param {Object} data Additional that is passed to the handler.
         * @returns {Function} The handler proxy function.
         */
        addEvent: function (el, type, handler, data) {
            var self = this,
                    types = type.split(' '),
                    listener = function (e) {
                        return handler.call(self, this, e, data);
                    };
            for (var i = 0; i < types.length; i++) {
                var _type = types[i];
                if (el.attachEvent) {
                    el.attachEvent('on' + _type, listener);
                } else {
                    el.addEventListener(_type, listener);
                }
            }
            this.eventHandlers.push({
                el: el,
                type: type,
                handler: listener
            });
            return listener;
        },
        /**
         * Removes an event listener.
         * 
         * @param {Element} el The element
         * @param {String} type Event type. Can be a space separated list of types.
         * @param {Function} handler The handler function
         * @returns {undefined}
         */
        removeEvent: function (el, type, handler) {
            var types = type.split(' ');
            for (var i = 0; i < types.length; i++) {
                var _type = types[i];
                if (el.detachEvent) {
                    el.detachEvent('on' + _type, handler);
                } else {
                    el.removeEventListener(_type, handler);
                }
            }
        },
        live: function (elClass, type, cb, context, data) {
            context = context || document;
            var self = this,
                    listener = function (el2, ev) {
                        var found, el = ev.target || ev.srcElement;
                        while (el && !(found = self.hasClass(el, elClass))) {
                            el = el.parentElement;
                        }
                        if (found) {
                            cb.apply(this, [el, ev, data]);
                        }
                    };
            return this.addEvent(context, type, listener, data);
        },
        destroy: function () {
            var instances = this.instances,
                    eventHandlers = this.eventHandlers;

            for (var i = 0; i < eventHandlers.length; i++) {
                var eventHandler = eventHandlers[i];
                this.removeEvent(eventHandler.el, eventHandler.type, eventHandler.handler);
            }

            for (var k = 0; k < instances.length; k++) {
                var instance = instances[k],
                        textField = instance.textField;
                if (instance.autocompleteAttr) {
                    textField.setAttribute('autocomplete', instance.autocompleteAttr);
                } else {
                    textField.removeAttribute('autocomplete');
                }
                instance.suggestionsContainer.parentNode.removeChild(
                        instance.suggestionsContainer);
                delete textField.autoCompleteInstance;
            }
        },
        onResize: function (el, ev, data) {
            var instances = this.instances;
            for (var i = 0; i < instances.length; i++) {
                this.updateSuggestionsContainer(instances[i], true);
            }
        },
        onMouseLeave: function (el, ev, data) {
            var suggestionsContainer = data.instance.suggestionsContainer;
            var selected = suggestionsContainer.querySelector('.autocomplete-suggestion.selected');
            if (selected) {
                setTimeout(function () {
                    selected.className = selected.className.replace(/\s*selected/, '');
                }, 20);
            }
        },
        onMouseOver: function (el, ev, data) {
            var sc = data.instance.suggestionsContainer,
                    selected = sc.querySelector('.autocomplete-suggestion.selected');
            if (selected) {
                selected.className = selected.className.replace(/\s*selected/, '');
            }
            el.className += ' selected';
        },
        onMouseDown: function (el, ev, data) {
            var instance = data.instance,
                    textField = instance.textField;
            if (this.hasClass(el, 'autocomplete-suggestion')) { // else outside click
                var v = el.getAttribute('data-val');
                textField.value = v;
                this.settings.onSelect(ev, v, el);
                this.hide(instance);
            }
        },
        onBlur: function (el, ev, data) {
            var instance = data.instance,
                    sc = instance.suggestionsContainer,
                    overSb = 0,
                    self = this;
            try {
                overSb = document.querySelector('.autocomplete-suggestions:hover');
            } catch (e) {
                // Default value: overSb = 0
            }

            if (!overSb) {
                instance.lastVal = el.value;
                self.hide(instance);
                setTimeout(function () {
                    self.hide(instance);
                }, 350); // hide suggestions on fast input
            } else if (el !== document.activeElement) {
                setTimeout(function () {
                    el.focus();
                }, 20);
            }
        },
        onKeyDown: function (el, ev, data) {
            var key = window.event ? ev.keyCode : ev.which,
                    instance = data.instance,
                    sc = instance.suggestionsContainer,
                    selected,
                    self = this;
            // down (40), up (38)
            if ((key === 40 || key === 38) && sc.innerHTML) {
                var next;
                selected = sc.querySelector('.autocomplete-suggestion.selected');
                if (!selected) {
                    next = (key === 40) ?
                            sc.querySelector('.autocomplete-suggestion') :
                            sc.childNodes[sc.childNodes.length - 1]; // first : last
                    next.className += ' selected';
                    el.value = next.getAttribute('data-val');
                } else {
                    next = (key === 40) ? selected.nextSibling : selected.previousSibling;
                    if (next) {
                        selected.className = selected.className.replace(/\s*selected/, '');
                        next.className += ' selected';
                        el.value = next.getAttribute('data-val');
                    } else {
                        selected.className = selected.className.replace(/\s*selected/, '');
                        el.value = instance.lastVal;
                        next = 0;
                    }
                }
                this.updateSuggestionsContainer(instance, false, next);
                return false;
            } else if (key === 27) { // esc
                el.value = instance.lastVal;
                this.hide(instance);
            } else if (key === 13 || key === 9) { // enter
                selected = sc.querySelector('.autocomplete-suggestion.selected');
                if (selected && this.isVisible(instance)) {
                    this.settings.onSelect(ev, selected.getAttribute('data-val'), selected);
                    setTimeout(function () {
                        self.hide(instance);
                    }, 20);
                }
            }
        },
        onKeyUp: function (el, ev, data) {
            var key = window.event ? ev.keyCode : ev.which,
                    self = this,
                    settings = self.settings,
                    instance = data.instance;
            if (!key || (key < 35 || key > 40) && key !== 13 && key !== 27) {
                var val = instance.textField.value;
                if (val.length >= settings.minChars) {
                    if (val !== instance.lastVal) {
                        instance.lastVal = val;
                        clearTimeout(instance.timer);
                        var suggestProxy = function (suggestions) {
                            self.suggest(instance, val, suggestions);
                        };
                        if (settings.cache) {
                            if (val in instance.cache) {
                                suggestProxy(instance.cache[val]);
                                return;
                            }
                            // no requests if previous suggestions were empty
                            for (var i = 1; i < val.length - settings.minChars; i++) {
                                var part = val.slice(0, val.length - i);
                                if (part in instance.cache && !instance.cache[part].length) {
                                    suggestProxy([]);
                                    return;
                                }
                            }
                        }
                        instance.timer = setTimeout(function () {
                            settings.source(val, suggestProxy);
                        }, settings.delay);
                    }
                } else {
                    instance.lastVal = val;
                    this.hide(instance);
                }
            }
        },
        onFocus: function (el, ev, data) {
            data.instance.lastVal = '\n';
            this.onKeyUp.call(this, el, ev, data);
        }
    };

    return autoComplete;

}));

/**
 * 
 * @constructor
 * @class eu_maxschuster_vaadin_autocompletetextfield_AutocompleteTextFieldExtension
 */
function eu_maxschuster_vaadin_autocompletetextfield_AutocompleteTextFieldExtension() {

    /* jshint validthis:true */
    "use strict";
    
    var AutoComplete = window.autoComplete;
    
    function CustomAutoComplete() {
        AutoComplete.apply(this, arguments);
    }
    
    CustomAutoComplete.prototype = Object.create(AutoComplete.prototype);
    
    CustomAutoComplete.prototype.updateSuggestionsContainer = function(instance, resize, next) {
        AutoComplete.prototype.updateSuggestionsContainer.apply(this, arguments);
        var textField = instance.textField,
                sc = instance.suggestionsContainer,
                scStyle = sc.style,
                found = false,
                zIndexParent = textField;
        // check if the textfield is inside another overlay
        while (zIndexParent && zIndexParent.parentElement && 
                !(found = this.hasClass(zIndexParent.parentElement, 'v-overlay-container'))) {
            zIndexParent = zIndexParent.parentElement;
        }
        if (found) {
            var style = this.getStyle(zIndexParent),
                    zIndex = style.zIndex || 0,
                    newZIndex = parseInt(zIndex, 10) + 1;
            // Adjust the suggestioncontainers z-index to be higher than
            // the z-index of the textfield overlay parent.
            scStyle.zIndex = newZIndex;
        } else if (scStyle.zIndex) {
            // Reset the ajusted z-index. Not sure if this is necessary... ^^
            scStyle.zIndex = "";
        }
    };

    var self = this,
            fontIconPrefix = "fonticon://";

    this.init = function () {
        this.lastResponseId = 0;
        this.pendingResponses = {};
        this.textField = this.findTextField();
        this.popupContainer = this.findPopupContainer();
        this.autoComplete = this.createAutoComplete(this.getConfig(this.getState()));
        this.scrollBehavior = null;
        this.scrollListener = false;
    };

    /**
     * Extends and extends the parent component of this extension.
     * @returns {Element} The parent element.
     */
    this.findTextField = function () {
        var textField = this.getElement(this.getParentId());
        return textField;
    };

    this.createAutoComplete = function (config) {
        var autoComplete = new CustomAutoComplete(config);
        this.currentConfig = config;
        return autoComplete;
    };

    this.findPopupContainer = function () {
        return document.querySelector('.v-overlay-container');
    };

    this.getConfig = function (state) {
        var menuClass = this.isArray(state.menuStyleNames) ?
                state.menuStyleNames.join(" ") : "";
        return {
            selector: this.textField,
            source: this.source,
            minChars: state.minChars,
            delay: state.delay,
            cache: state.cache,
            menuClass: menuClass,
            renderItem: this.renderItem,
            onSelect: this.onSelect,
            popupContainer: this.popupContainer
        };
    };

    this.compareConfig = function (oldConfig, newConfig) {
        for (var key in oldConfig) {
            if (oldConfig[key] !== newConfig[key]) {
                return false;
            }
        }
        return true;
    };

    this.onSelect = function (event, value, item) {
        var textField = self.textField;
        // Fake keydown to force a text change event
        self.triggerEvent(textField, "keydown");
    };

    this.onStateChange = function () {
        var state = this.getState(),
                oldConfig = this.currentConfig,
                newConfig = this.getConfig(state),
                autoComplete = this.autoComplete;
        if (!this.compareConfig(oldConfig, newConfig)) {
            if (typeof autoComplete === "object") {
                autoComplete.destroy();
            }
            this.autoComplete = this.createAutoComplete(newConfig);
        }
        this.setScrollBehavoir(state.scrollBehavior);
    };

    this.setScrollBehavoir = function (scrollBehavior) {
        switch (scrollBehavior) {
            case "REFRESH":
            case "CLOSE":
                if (!this.scrollListener) {
                    this.addEvent(window, "scroll", this.onScroll, true);
                    this.scrollListener = true;
                }
                break;
            default: // including "NONE"
                if (this.scrollListener) {
                    this.removeEvent(window, "scroll", this.onScroll, true);
                    this.scrollListener = false;
                }
                break;
        }
        this.scrollBehavior = scrollBehavior;
    };

    this.onScroll = function (ev) {
        var scrollBehavior = self.scrollBehavior,
                autoComplete = self.autoComplete,
                autoCompleteInstance = self.textField.autoCompleteInstance,
                suggestionsContainer = autoCompleteInstance.suggestionsContainer;
        
        if (!autoComplete.isVisible(autoCompleteInstance)) {
            return;
        }
        
        // ignore scroll events from inside the suggestionsContainer
        var found, el = ev.target || ev.srcElement;
        while (el && !(found = (el === suggestionsContainer))) {
            el = el.parentElement;
        }
        if (found) {
            return;
        }
        
        switch (scrollBehavior) {
            case "REFRESH":
                autoComplete.updateSuggestionsContainer(autoCompleteInstance, true);
                break;
            case "CLOSE":
                autoComplete.hide(autoCompleteInstance);
                break;
            default:
                // Do nothing
                break;
        }
    };

    this.onUnregister = function () {
        this.autoComplete.destroy();
        if (this.scrollListener) {
            this.removeEvent(window, "scroll", this.onScroll, true);
        }
    };

    this.setSuggestions = function (responseId, suggestions) {
        var pendingResponses = this.pendingResponses;
        if (typeof pendingResponses[responseId] === "function") {
            pendingResponses[responseId](suggestions);
            delete pendingResponses[responseId];
        }
    };

    /**
     * @param {type} unsafe
     * @returns {String}
     * @see http://stackoverflow.com/a/6234804
     */
    this.escapeHtml = function (unsafe) {
        if (!unsafe || typeof unsafe !== "string") {
            // Is no string or empty. Do nothing with it...
            return unsafe;
        }
        return unsafe.replace(/&/g, "&amp;")
                .replace(/</g, "&lt;")
                .replace(/>/g, "&gt;")
                .replace(/"/g, "&quot;")
                .replace(/'/g, "&#039;");
    };

    /**
     * Cross browser add event.
     * 
     * @param {Element} ob
     * @param {string} type
     * @param {Function} fn
     * @param {Boolean} useCapture
     * @returns {undefined}
     */
    this.addEvent = function (ob, type, fn, useCapture) {
        if (ob.addEventListener) {
            ob.addEventListener(type, fn, useCapture || false);
        } else if (ob.attachEvent) {
            ob.attachEvent('on' + type, fn);
        }
    };

    this.removeEvent = function (ob, type, fn, useCapture) {
        if (ob.removeEventListener) {
            ob.removeEventListener(type, fn, useCapture || false);
        } else if (ob.dettachEvent) {
            ob.dettachEvent('on' + type, fn);
        }
    };

    /**
     * Trigger an Event
     * 
     * @param {Element} element
     * @param {string} event
     * @returns {undefined}
     * @see http://stackoverflow.com/a/2856602
     */
    this.triggerEvent = function (element, event) {
        if ("createEvent" in document) {
            var evt = document.createEvent("HTMLEvents");
            evt.initEvent(event, false, true);
            element.dispatchEvent(evt);
        } else {
            element.fireEvent("on" + event);
        }
    };

    this.source = function (term, response) {
        var responseId = ++self.lastResponseId;
        response.term = term;
        self.serverQuerySuggestions(responseId, term);
        self.pendingResponses[responseId] = response;
    };

    this.renderItem = function (item, search) {
        var escape = self.getState().itemsAsHtml,
                valueEscaped = self.escapeHtml(item.value),
                value = item.value,
                description = item.description,
                icon = item.icon,
                itemClass = 'autocomplete-suggestion',
                classes = [itemClass],
                styleNames = item.styleNames;

        if (escape) {
            value = valueEscaped;
            description = self.escapeHtml(description);
        }

        if (description) {
            classes.push('has-description');
        }

        if (icon) {
            classes.push('has-icon');
        }

        //  Add the item styles
        if (self.isArray(styleNames)) {
            classes.push.apply(classes, styleNames);
        }

        var rendered = '<div class="' + classes.join(' ') + '" data-val="' + valueEscaped + '">';
        rendered += '<div class="' + itemClass + '-content">';
        if (icon) {
            rendered += '<div class="' + itemClass + '-icon">';
            rendered += self.getIconHtml(icon);
            rendered += '</div>';
        }
        rendered += '<div class="' + itemClass + '-text">';
        rendered += '<div class="' + itemClass + '-value"><span>' +
                value +
                '</span></div>';
        if (description) {
            rendered += '<div class="' + itemClass + '-description"><span>' +
                    description +
                    '</span></div>';
        }
        rendered += '</div>'; // itemClass + '-value'
        rendered += '</div>'; // itemClass + '-text'
        rendered += '</div>'; // itemClass + '-content'

        return rendered;
    };

    this.getResource = function (resourceKey) {
        var resources = this.getState().resources;
        if (typeof resources[resourceKey] !== "object") {
            return null;
        }
        return resources[resourceKey];
    };

    this.getIconHtml = function (resourceKey) {
        var resource = this.getResource(resourceKey);
        if (typeof resource !== "object") {
            return false;
        }
        var vaadinUri = resource.uRL;
        if (vaadinUri.substr(0, fontIconPrefix.length) === fontIconPrefix) {
            // is fonticon
            var parts = vaadinUri.substr(fontIconPrefix.length).split("/"),
                    fontFamily = decodeURI(parts[0]),
                    codepoint = String.fromCharCode(parseInt(parts[1], 16));
            return '<span class="v-icon ' + fontFamily + '">' +
                    codepoint +
                    '</span>';
        } else {
            return '<img class="v-icon" src="' + this.translateVaadinUri(vaadinUri) + '" />';
        }
    };

    /**
     * Checks if the given object is an array.
     * @param {Object} object
     * @returns {Boolean}
     */
    this.isArray = function (object) {
        if (Array.isArray) {
            return Array.isArray(object);
        }
        return Object.prototype.toString.call(object) === '[object Array]';
    };

    this.init();

}